// Autogenerated code; DO NOT EDIT.

/*
 * Schema Open API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type NumaNodeProcessor struct {
	TotalAssignedProcessors  uint32 `json:"TotalAssignedProcessors,omitempty"`
	TotalAvailableProcessors uint32 `json:"TotalAvailableProcessors,omitempty"`
}
