// Package sshsig provides an API to sign and verify messages using SSH keys.
// It is an implementation of the SSH Signature format as described in
// https://github.com/openssh/openssh-portable/blob/V_9_3_P1/PROTOCOL.sshsig.
package sshsig
