// Autogenerated code; DO NOT EDIT.

/*
 * Schema Open API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swaggerapi/swaggercodegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swaggerapi/swaggercodegen.git)
 */

package hcsschema

type MemoryBackingType string

// List of MemoryBackingType
const (
	MemoryBackingType_PHYSICAL MemoryBackingType = "Physical"
	MemoryBackingType_VIRTUAL  MemoryBackingType = "Virtual"
	MemoryBackingType_HYBRID   MemoryBackingType = "Hybrid"
)
